from typing import Dict

import requests
from google.auth.transport.requests import Request
from google.oauth2 import service_account

from gcp_pilot.base import GoogleCloudPilotAPI


class IdentityAwareProxy(GoogleCloudPilotAPI):
    def __init__(self, audience: str):
        self.audience = audience
        super().__init__()

    def _build_client(self, **kwargs) -> None:
        return None

    def _get_id_credentials(self):
        credentials = service_account.IDTokenCredentials(
            signer=self.credentials.signer,
            service_account_email=self.credentials.service_account_email,
            token_uri="https://oauth2.googleapis.com/token",  # TODO: fetch from credentials
            target_audience=self.audience,
        )
        if not credentials.valid:
            credentials.refresh(request=Request())
        return credentials

    def _get_gcp_token(self):
        email = self.credentials.service_account_email
        metadata_server_token_url = f"http://metadata/computeMetadata/v1/instance/service-accounts/{email}/identity"

        token_request_url = f"{metadata_server_token_url}?audience={self.audience}"
        token_request_headers = {"Metadata-Flavor": "Google"}

        token_response = requests.get(url=token_request_url, headers=token_request_headers)
        return token_response.content.decode("utf-8")

    @property
    def token(self) -> str:
        try:
            return self._get_gcp_token()
        except requests.exceptions.ConnectionError:
            credentials = self._get_id_credentials()
            return credentials.token

    @property
    def headers(self) -> Dict[str, str]:
        return {"Authorization": f"Bearer {self.token}"}


__all__ = ("IdentityAwareProxy",)



# More Information: <https://googleapis.dev/python/clouderrorreporting/latest/usage.html>
from google.cloud import error_reporting
from google.cloud.error_reporting import HTTPContext

from gcp_pilot.base import GoogleCloudPilotAPI, DiscoveryMixin


class CloudErrorReporting(GoogleCloudPilotAPI):
    _client_class = error_reporting.Client

    def __init__(self, service_name):
        super().__init__(service=service_name)

    def report(self, message, http_context=None, user=None) -> None:
        self.client.report(message=message, http_context=http_context, user=user)

    def report_with_request(self, request: "WSGIRequest", status_code: int, message: str = None) -> None:
        user = str(request.user) if request.user.is_authenticated else None
        http_context = self._wsgi_to_http_context(request=request, status_code=status_code)

        if not message:
            self.client.report_exception(
                http_context=http_context,
                user=user,
            )
        else:
            self.report(
                message=message,
                http_context=http_context,
                user=user,
            )

    def _wsgi_to_http_context(self, request: "WSGIRequest", status_code: int) -> HTTPContext:
        return HTTPContext(
            url=request.get_raw_uri(),
            method=request.method,
            user_agent=request.headers.get("User-Agent"),
            referrer=request.headers.get("Referer"),
            response_status_code=status_code,
        )


class CloudErrorExplorer(DiscoveryMixin, GoogleCloudPilotAPI):
    def __init__(self):
        super().__init__(
            serviceName="clouderrorreporting",
            version="v1beta1",
            cache_discovery=False,
        )

    def get_events(
        self,
        error_id: str = None,
        service_name: str = None,
        service_version: str = None,
        resource_type: str = None,
        project_id: str = None,
    ):
        params = dict(
            projectName=self._project_path(project_id=project_id),
            groupId=error_id,
        )
        if service_name:
            params["serviceFilter_service"] = service_name
        if service_version:
            params["serviceFilter_version"] = service_version
        if resource_type:
            params["serviceFilter_resourceType"] = resource_type

        yield from self._paginate(
            method=self.client.projects().events().list,
            params=params,
            result_key="errorEvents",
        )

    def get_errors(
        self,
        service_name: str = None,
        service_version: str = None,
        resource_type: str = None,
        project_id: str = None,
    ):
        params = dict(
            projectName=self._project_path(project_id=project_id),
        )
        if service_name:
            params["serviceFilter_service"] = service_name
        if service_version:
            params["serviceFilter_version"] = service_version
        if resource_type:
            params["serviceFilter_resourceType"] = resource_type

        items = self._paginate(
            method=self.client.projects().groupStats().list,
            params=params,
            result_key="errorGroupStats",
        )
        for item in items:
            yield {"id": item["group"]["groupId"], **item}


__all__ = (
    "CloudErrorReporting",
    "CloudErrorExplorer",
)



# More Information: https://googleapis.dev/python/storage/latest/index.html
import io
import os
from typing import AsyncIterator

import requests
from google.cloud import storage
from google.cloud.exceptions import Conflict
from google.cloud.storage import Bucket, Blob

from gcp_pilot.base import GoogleCloudPilotAPI


class CloudStorage(GoogleCloudPilotAPI):
    _client_class = storage.Client

    async def create_bucket(
        self,
        name: str,
        region: str = None,
        project_id: str = None,
        exists_ok: bool = True,
    ) -> Bucket:
        bucket = self.client.bucket(name)
        try:
            return self.client.create_bucket(
                bucket_or_name=bucket,
                location=region,
                project=project_id or self.project_id,
            )
        except Conflict:
            if not exists_ok:
                raise
            return await self.check_bucket(name=name)

    async def check_bucket(self, name: str) -> Bucket:
        return self.client.get_bucket(bucket_or_name=name)

    async def upload(
        self,
        source_file,
        bucket_name: str,
        target_file_name: str = None,
        chunk_size: int = None,
        project_id: str = None,
        region: str = None,
        is_public: bool = False,
    ) -> Blob:
        target_bucket = await self.create_bucket(name=bucket_name, project_id=project_id, region=region)

        target_file_name = target_file_name or str(source_file).rsplit("/", maxsplit=1)[-1]
        blob = target_bucket.blob(target_file_name, chunk_size=chunk_size)

        if isinstance(source_file, str):
            if source_file.startswith("http"):
                file_obj = self._download(url=source_file)
                blob.upload_from_file(file_obj)
            elif os.path.exists(source_file):
                blob.upload_from_filename(source_file)
            else:
                content = io.StringIO(source_file)
                blob.upload_from_file(content)
        else:
            blob.upload_from_file(source_file)

        if is_public:
            blob.make_public()

        return blob

    async def copy(
        self,
        source_file_name,
        source_bucket_name: str,
        target_bucket_name: str,
        target_file_name: str = None,
        project_id: str = None,
        region: str = None,
    ) -> Blob:
        source_bucket = await self.create_bucket(name=source_bucket_name, region=region, project_id=project_id)
        source_blob = source_bucket.blob(source_file_name)

        target_bucket = await self.create_bucket(name=target_bucket_name, region=region, project_id=project_id)
        target_file_name = target_file_name or str(source_file_name).rsplit("/", maxsplit=1)[-1]

        obj = source_bucket.copy_blob(source_blob, target_bucket, target_file_name)
        return obj

    async def move(
        self,
        source_file_name: str,
        source_bucket_name: str,
        target_bucket_name: str,
        target_file_name: str = None,
        project_id: str = None,
        region: str = None,
    ) -> Blob:
        data = await self.copy(
            source_file_name=source_file_name,
            source_bucket_name=source_bucket_name,
            target_file_name=target_file_name,
            target_bucket_name=target_bucket_name,
            project_id=project_id,
            region=region,
        )
        await self.delete(file_name=source_file_name, bucket_name=source_bucket_name)
        return data

    async def delete(self, file_name: str, bucket_name: str = None) -> None:
        bucket = await self.check_bucket(name=bucket_name)
        blob = bucket.blob(file_name)
        return blob.delete()

    async def list_files(self, bucket_name: str, prefix: str = None) -> AsyncIterator[Blob]:
        blobs = self.client.list_blobs(
            bucket_name,
            prefix=prefix,
        )
        for blob in blobs:
            yield blob

    def _download(self, url: str) -> io.BytesIO:
        response = requests.get(url, stream=True)
        file = io.BytesIO()
        file.write(response.content)
        file.seek(0)
        return file

    def get_uri(self, blob: Blob) -> str:
        return f"gs://{blob.bucket.name}/{blob.name}"


__all__ = ("CloudStorage",)



# More Information: https://developers.google.com/admin-sdk/directory/reference/rest
from typing import Dict, Any, Generator

from gcp_pilot.base import GoogleCloudPilotAPI, DiscoveryMixin

UserType = GroupType = MemberType = Dict[str, Any]


class Directory(DiscoveryMixin, GoogleCloudPilotAPI):
    _scopes = [
        "https://www.googleapis.com/auth/admin.directory.group",
        "https://www.googleapis.com/auth/admin.directory.user",
    ]

    def __init__(self, email: str, **kwargs):
        self.email = email
        super().__init__(
            serviceName="admin",
            version="directory_v1",
            cache_discovery=False,
            subject=email,
            **kwargs,
        )

    def _build_context(self, customer: str = None, domain: str = None) -> Dict[str, str]:
        context = {
            "customer": customer or "my_customer",
        }
        if domain:
            context["domain"] = domain
        return context

    def get_users(self, customer: str = None, domain: str = None) -> Generator[UserType, None, None]:
        params = self._build_context(customer=customer, domain=domain)
        yield from self._paginate(
            method=self.client.users().list,
            result_key="users",
            params=params,
            order_by="email",
        )

    def get_groups(self, customer: str = None, domain: str = None) -> Generator[GroupType, None, None]:
        params = self._build_context(customer=customer, domain=domain)
        yield from self._paginate(
            method=self.client.groups().list,
            result_key="groups",
            params=params,
        )

    def get_group(self, group_id: str) -> GroupType:
        return self._execute(
            method=self.client.groups().get,
            groupKey=group_id,
        )

    def create_or_update_group(
        self,
        email: str,
        name: str = None,
        description: str = None,
        group_id: str = None,
    ) -> GroupType:
        body = dict(
            email=email,
            name=name,
            description=description,
        )

        if not group_id:
            return self._execute(
                method=self.client.groups().insert,
                body=body,
            )

        return self._execute(
            method=self.client.groups().update,
            body=body,
        )

    def delete_group(self, group_id: str) -> GroupType:
        return self._execute(
            method=self.client.groups().delete,
            groupKey=group_id,
        )

    def get_group_members(self, group_id: str) -> Generator[MemberType, None, None]:
        params = dict(
            groupKey=group_id,
        )

        yield from self._paginate(
            method=self.client.members().list,
            result_key="members",
            params=params,
        )

    def add_group_member(self, group_id: str, email: str, role: str = "MEMBER") -> MemberType:
        body = {"email": email, "role": role}

        return self._execute(
            method=self.client.members().insert,
            groupKey=group_id,
            body=body,
        )

    def delete_group_member(self, group_id: str, member_id: str) -> MemberType:
        return self._execute(
            method=self.client.members().delete,
            groupKey=group_id,
            memberKey=member_id,
        )


__all__ = ("Directory",)
